/// @file
////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2016/17      Christian Lessig, Otto-von-Guericke Universitaet Magdeburg
///
////////////////////////////////////////////////////////////////////////////////////////////////////
///
///  module     : tutorial 4
///
///  author     : lessig@isg.cs.ovgu.de
///
///  project    : GPU Programming
///
///  description: CUDA basics
///
////////////////////////////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <algorithm>
#include <chrono>

#include "cuda_util.h"

typedef std::chrono::milliseconds TimeT;


////////////////////////////////////////////////////////////////////////////////////////////////////
// add constant to vector entries
////////////////////////////////////////////////////////////////////////////////////////////////////
__global__
void
vecAddC( float* data, const unsigned int n, const float val) {

  unsigned int tid = blockIdx.x * blockDim.x + threadIdx.x;

  if( tid < n) {
    for( unsigned int it = 0; it < 10000000; it++){
      data[tid] += val;
    }
  }
}


////////////////////////////////////////////////////////////////////////////////////////////////////
// program entry point
////////////////////////////////////////////////////////////////////////////////////////////////////
int
main( int /*argc*/, char** /*argv*/ ) {

  const unsigned int n =  1024;
  float* data1 = (float*) malloc( sizeof(float) * n);

  for( unsigned int i = 0; i < n; ++i) {
    data1[i] = (float) i;
  }

  // check execution environment
  int deviceCount = 0;
  checkErrorsCuda( cudaGetDeviceCount(&deviceCount));
  if( 0 == deviceCount) {
    std::cerr << "initDevice() : No CUDA device found." << std::endl;
    return EXIT_FAILURE;
  }

  // query the device properties
  cudaDeviceProp devProp;
  cudaGetDeviceProperties(&devProp, 0);
  printDeviceProps( devProp);

  // set the device
  int device_handle = 0;
  cudaSetDevice( device_handle);

  // initialize memory
  float* data_device1 = nullptr;
  // allocate device memory
  checkErrorsCuda( cudaMalloc((void **) &data_device1, sizeof(float) * n));
  // copy device memory
  checkErrorsCuda( cudaMemcpy( (void*) data_device1, data1, sizeof(float) * n, cudaMemcpyHostToDevice ));

  // determine thread layout
  const unsigned int MAX_THREADS_PER_BLOCK = 1024;
  int num_threads_per_block = std::min( n, MAX_THREADS_PER_BLOCK);
  int num_blocks = n / MAX_THREADS_PER_BLOCK;
  if( 0 != n % MAX_THREADS_PER_BLOCK) {
    num_blocks++;
  }
  std::cout << "Number of devices = " << deviceCount << std::endl;

  std::cout << "num_blocks = " << num_blocks << " :: "
            << "num_threads_per_block = " << num_threads_per_block << std::endl;

  std::chrono::system_clock::time_point start = std::chrono::system_clock::now();

  // run kernel
  vecAddC<<< num_blocks , num_threads_per_block >>>( data_device1, n, 1);
  checkLastCudaError("Kernel execution failed");
  cudaDeviceSynchronize();

  // copy result back to host
  checkErrorsCuda( cudaMemcpy( data1, data_device1, sizeof(float) * n, cudaMemcpyDeviceToHost ));

  double time = std::chrono::duration_cast<TimeT>(std::chrono::system_clock::now()-start).count();
  std::cout<< "time: " << time <<" ms"<<std::endl;

  // clean up device memory
  checkErrorsCuda( cudaFree( data_device1));
  // clean up host memory
  free( data1);

  return EXIT_SUCCESS;
}
